@@ -1,86 +1,149 @@
 # -*- coding: utf-8 -*-
 """
-Import raw data, annotations, and spike sorting results as Neo objects
+The :mod:`neurotic.datasets.data` module implements a function for loading a
+dataset from selected metadata.
+
+.. autofunction:: load_dataset
 """
 
+from packaging import version
 import numpy as np
 import pandas as pd
 import quantities as pq
 import elephant
 import neo
 
-from ..datasets.metadata import abs_path
-from neo.test.generate_datasets import fake_neo
+from ..datasets.metadata import _abs_path
 
-def LoadAndPrepareData(metadata, lazy=False, signal_group_mode='split-all', fake_data_for_testing = False):
+def load_dataset(metadata, lazy=False, signal_group_mode='split-all', filter_events_from_epochs=False):
     """
+    Load a dataset.
 
+    ``metadata`` may be a :class:`MetadataSelector
+    <neurotic.datasets.metadata.MetadataSelector>` or a simple dictionary
+    containing the appropriate data.
+
+    The ``data_file`` in ``metadata`` is read into a Neo :class:`Block
+    <neo.core.Block>` using an automatically detected :mod:`neo.io` class
+    if ``lazy=False`` or a :mod:`neo.rawio` class if ``lazy=True``.
+
+    Epochs and events loaded from ``annotations_file`` and
+    ``epoch_encoder_file`` and spike trains loaded from ``tridesclous_file``
+    are added to the Neo Block.
+
+    If ``lazy=False``, filters given in ``metadata`` are applied to the
+    signals and amplitude discriminators are run to detect spikes.
     """
 
     # read in the electrophysiology data
-    blk = ReadDataFile(metadata, lazy, signal_group_mode)
-    # blk = CreateNeoBlockExample()
+    blk = _read_data_file(metadata, lazy, signal_group_mode)
 
     # apply filters to signals if not using lazy loading of signals
     if not lazy:
-        blk = ApplyFilters(metadata, blk)
+        blk = _apply_filters(metadata, blk)
+
+    # copy events into epochs and vice versa
+    epochs_from_events = [neo.Epoch(name=ev.name, times=ev.times, labels=ev.labels, durations=np.zeros_like(ev.times)) for ev in blk.segments[0].events]
+    events_from_epochs = [neo.Event(name=ep.name, times=ep.times, labels=ep.labels) for ep in blk.segments[0].epochs]
+    if not filter_events_from_epochs:
+        blk.segments[0].epochs += epochs_from_events
+    blk.segments[0].events += events_from_epochs
 
     # read in annotations
-    annotations_dataframe = ReadAnnotationsFile(metadata)
-    blk.segments[0].epochs += CreateNeoEpochsFromDataframe(annotations_dataframe, metadata, abs_path(metadata, 'annotations_file'))
-    blk.segments[0].events += CreateNeoEventsFromDataframe(annotations_dataframe, metadata, abs_path(metadata, 'annotations_file'))
+    annotations_dataframe = _read_annotations_file(metadata)
+    blk.segments[0].epochs += _create_neo_epochs_from_dataframe(annotations_dataframe, metadata, _abs_path(metadata, 'annotations_file'), filter_events_from_epochs)
+    blk.segments[0].events += _create_neo_events_from_dataframe(annotations_dataframe, metadata, _abs_path(metadata, 'annotations_file'))
 
     # read in epoch encoder file
-    epoch_encoder_dataframe = ReadEpochEncoderFile(metadata)
-    blk.segments[0].epochs += CreateNeoEpochsFromDataframe(epoch_encoder_dataframe, metadata, abs_path(metadata, 'epoch_encoder_file'))
-    blk.segments[0].events += CreateNeoEventsFromDataframe(epoch_encoder_dataframe, metadata, abs_path(metadata, 'epoch_encoder_file'))
+    epoch_encoder_dataframe = _read_epoch_encoder_file(metadata)
+    blk.segments[0].epochs += _create_neo_epochs_from_dataframe(epoch_encoder_dataframe, metadata, _abs_path(metadata, 'epoch_encoder_file'), filter_events_from_epochs)
+    blk.segments[0].events += _create_neo_events_from_dataframe(epoch_encoder_dataframe, metadata, _abs_path(metadata, 'epoch_encoder_file'))
 
     # classify spikes by amplitude if not using lazy loading of signals
     if not lazy:
-        blk.segments[0].spiketrains += RunAmplitudeDiscriminators(metadata, blk)
+        blk.segments[0].spiketrains += _run_amplitude_discriminators(metadata, blk)
 
     # read in spikes identified by spike sorting using tridesclous
     t_start = blk.segments[0].analogsignals[0].t_start
     t_stop = blk.segments[0].analogsignals[0].t_stop
     sampling_period = blk.segments[0].analogsignals[0].sampling_period
-    spikes_dataframe = ReadSpikesFile(metadata, blk)
-    blk.segments[0].spiketrains += CreateNeoSpikeTrainsFromDataframe(spikes_dataframe, metadata, t_start, t_stop, sampling_period)
+    spikes_dataframe = _read_spikes_file(metadata, blk)
+    blk.segments[0].spiketrains += _create_neo_spike_trains_from_dataframe(spikes_dataframe, metadata, t_start, t_stop, sampling_period)
 
-    if fake_data_for_testing:
-        # load fake data as a demo
-        blk.segments[0].epochs += [fake_neo('Epoch') for _ in range(5)]
-        blk.segments[0].events += [fake_neo('Event') for _ in range(5)]
-        blk.segments[0].spiketrains += [fake_neo('SpikeTrain') for _ in range(5)]
+    # alphabetize epoch and event channels by name
+    blk.segments[0].epochs.sort(key=lambda ep: ep.name)
+    blk.segments[0].events.sort(key=lambda ev: ev.name)
 
     return blk
 
-def ReadDataFile(metadata, lazy=False, signal_group_mode='split-all'):
+def _read_data_file(metadata, lazy=False, signal_group_mode='split-all'):
     """
-
+    Read in the ``data_file`` given in ``metadata`` using an automatically
+    detected :mod:`neo.io` class if ``lazy=False`` or a :mod:`neo.rawio` class
+    if ``lazy=True``. If ``lazy=True``, manually load epochs, events, and spike
+    trains, but not signals. Return a Neo :class:`Block <neo.core.Block>`.
     """
 
     # read in the electrophysiology data
     # - signal_group_mode='split-all' ensures every channel gets its own
     #   AnalogSignal, which is important for indexing in EphyviewerConfigurator
-    io = neo.io.get_io(abs_path(metadata, 'data_file'))
+    io = neo.io.get_io(_abs_path(metadata, 'data_file'))
     blk = io.read_block(lazy=lazy, signal_group_mode=signal_group_mode)
 
-    # load all proxy objects except analog signals
+    # load all objects except analog signals
     if lazy:
-        for i in range(len(blk.segments[0].epochs)):
-            epoch = blk.segments[0].epochs[i]
-            if hasattr(epoch, 'load'):
-                blk.segments[0].epochs[i] = epoch.load()
 
-        for i in range(len(blk.segments[0].events)):
-            event = blk.segments[0].events[i]
-            if hasattr(event, 'load'):
-                blk.segments[0].events[i] = event.load()
+        if version.parse(neo.__version__) >= version.parse('0.8.0.dev'):  # Neo >= 0.8.0 has proxy objects with load method
 
-        for i in range(len(blk.segments[0].spiketrains)):
-            spiketrain = blk.segments[0].spiketrains[i]
-            if hasattr(spiketrain, 'load'):
-                blk.segments[0].spiketrains[i] = spiketrain.load()
+            for i in range(len(blk.segments[0].epochs)):
+                epoch = blk.segments[0].epochs[i]
+                if hasattr(epoch, 'load'):
+                    blk.segments[0].epochs[i] = epoch.load()
+
+            for i in range(len(blk.segments[0].events)):
+                event = blk.segments[0].events[i]
+                if hasattr(event, 'load'):
+                    blk.segments[0].events[i] = event.load()
+
+            for i in range(len(blk.segments[0].spiketrains)):
+                spiketrain = blk.segments[0].spiketrains[i]
+                if hasattr(spiketrain, 'load'):
+                    blk.segments[0].spiketrains[i] = spiketrain.load()
+
+        else:  # Neo < 0.8.0 does not have proxy objects
+
+            neorawioclass = neo.rawio.get_rawio_class(_abs_path(metadata, 'data_file'))
+            if neorawioclass is not None:
+                neorawio = neorawioclass(_abs_path(metadata, 'data_file'))
+                neorawio.parse_header()
+
+                for i in range(len(blk.segments[0].epochs)):
+                    epoch = blk.segments[0].epochs[i]
+                    channel_index = next((i for i, chan in enumerate(neorawio.header['event_channels']) if chan['name'] == epoch.name and chan['type'] == b'epoch'), None)
+                    if channel_index is not None:
+                        ep_raw_times, ep_raw_durations, ep_labels = neorawio.get_event_timestamps(event_channel_index=channel_index)
+                        ep_times = neorawio.rescale_event_timestamp(ep_raw_times, dtype='float64')
+                        ep_durations = neorawio.rescale_epoch_duration(ep_raw_durations, dtype='float64')
+                        ep = neo.Epoch(times=ep_times*pq.s, durations=ep_durations*pq.s, labels=ep_labels, name=epoch.name)
+                        blk.segments[0].epochs[i] = ep
+
+                for i in range(len(blk.segments[0].events)):
+                    event = blk.segments[0].events[i]
+                    channel_index = next((i for i, chan in enumerate(neorawio.header['event_channels']) if chan['name'] == event.name and chan['type'] == b'event'), None)
+                    if channel_index is not None:
+                        ev_raw_times, _, ev_labels = neorawio.get_event_timestamps(event_channel_index=channel_index)
+                        ev_times = neorawio.rescale_event_timestamp(ev_raw_times, dtype='float64')
+                        ev = neo.Event(times=ev_times*pq.s, labels=ev_labels, name=event.name)
+                        blk.segments[0].events[i] = ev
+
+                for i in range(len(blk.segments[0].spiketrains)):
+                    spiketrain = blk.segments[0].spiketrains[i]
+                    channel_index = next((i for i, chan in enumerate(neorawio.header['unit_channels']) if chan['name'] == spiketrain.name), None)
+                    if channel_index is not None:
+                        st_raw_times = neorawio.get_spike_timestamps(unit_index=channel_index)
+                        st_times = neorawio.rescale_spike_timestamp(st_raw_times, dtype='float64')
+                        st = neo.SpikeTrain(times=st_times*pq.s, name=st.name)
+                        blk.segments[0].spiketrains[i] = st
 
     # convert byte labels to Unicode strings
     for epoch in blk.segments[0].epochs:
@@ -91,9 +154,10 @@ def ReadDataFile(metadata, lazy=False, signal_group_mode='split-all'):
 
     return blk
 
-def ReadAnnotationsFile(metadata):
+def _read_annotations_file(metadata):
     """
-
+    Read in epochs and events from the ``annotations_file`` in ``metadata`` and
+    return a dataframe.
     """
 
     if metadata['annotations_file'] is None:
@@ -111,7 +175,7 @@ def ReadAnnotationsFile(metadata):
         }
 
         # parse the file and create a dataframe
-        df = pd.read_csv(abs_path(metadata, 'annotations_file'), dtype = dtypes)
+        df = pd.read_csv(_abs_path(metadata, 'annotations_file'), dtype = dtypes)
 
         # increment row labels by 2 so they match the source file
         # which is 1-indexed and has a header
@@ -154,9 +218,10 @@ def ReadAnnotationsFile(metadata):
         # return the dataframe
         return df
 
-def ReadEpochEncoderFile(metadata):
+def _read_epoch_encoder_file(metadata):
     """
-
+    Read in epochs from the ``epoch_encoder_file`` in ``metadata`` and return a
+    dataframe.
     """
 
     if metadata['epoch_encoder_file'] is None:
@@ -173,7 +238,7 @@ def ReadEpochEncoderFile(metadata):
         }
 
         # parse the file and create a dataframe
-        df = pd.read_csv(abs_path(metadata, 'epoch_encoder_file'), dtype = dtypes)
+        df = pd.read_csv(_abs_path(metadata, 'epoch_encoder_file'), dtype = dtypes)
 
         # increment row labels by 2 so they match the source file
         # which is 1-indexed and has a header
@@ -222,9 +287,10 @@ def ReadEpochEncoderFile(metadata):
         # return the dataframe
         return df
 
-def ReadSpikesFile(metadata, blk):
+def _read_spikes_file(metadata, blk):
     """
-    Read in spikes identified by spike sorting with tridesclous.
+    Read in spikes identified by spike sorting with tridesclous and return a
+    dataframe.
     """
 
     if metadata['tridesclous_file'] is None or metadata['tridesclous_channels'] is None:
@@ -234,7 +300,7 @@ def ReadSpikesFile(metadata, blk):
     else:
 
         # parse the file and create a dataframe
-        df = pd.read_csv(abs_path(metadata, 'tridesclous_file'), names = ['index', 'label'])
+        df = pd.read_csv(_abs_path(metadata, 'tridesclous_file'), names = ['index', 'label'])
 
         # drop clusters with negative labels
         df = df[df['label'] >= 0]
@@ -251,17 +317,19 @@ def ReadSpikesFile(metadata, blk):
         # return the dataframe
         return df
 
-def CreateNeoEpochsFromDataframe(dataframe, metadata, file_origin):
+def _create_neo_epochs_from_dataframe(dataframe, metadata, file_origin, filter_events_from_epochs=False):
     """
-
+    Convert the contents of a dataframe into Neo :class:`Epochs
+    <neo.core.Epoch>`.
     """
 
     epochs_list = []
 
     if dataframe is not None:
 
-        # keep only rows with a positive duration
-        dataframe = dataframe[dataframe['Duration (s)'] > 0]
+        if filter_events_from_epochs:
+            # keep only rows with a positive duration
+            dataframe = dataframe[dataframe['Duration (s)'] > 0]
 
         # group epochs by type
         for type_name, df in dataframe.groupby('Type'):
@@ -280,9 +348,10 @@ def CreateNeoEpochsFromDataframe(dataframe, metadata, file_origin):
     # return the list of Neo Epochs
     return epochs_list
 
-def CreateNeoEventsFromDataframe(dataframe, metadata, file_origin):
+def _create_neo_events_from_dataframe(dataframe, metadata, file_origin):
     """
-
+    Convert the contents of a dataframe into Neo :class:`Events
+    <neo.core.Event>`.
     """
 
     events_list = []
@@ -305,9 +374,10 @@ def CreateNeoEventsFromDataframe(dataframe, metadata, file_origin):
     # return the list of Neo Events
     return events_list
 
-def CreateNeoSpikeTrainsFromDataframe(dataframe, metadata, t_start, t_stop, sampling_period):
+def _create_neo_spike_trains_from_dataframe(dataframe, metadata, t_start, t_stop, sampling_period):
     """
-
+    Convert the contents of a dataframe into Neo :class:`SpikeTrains
+    <neo.core.SpikeTrain>`.
     """
 
     spiketrain_list = []
@@ -323,7 +393,7 @@ def CreateNeoSpikeTrainsFromDataframe(dataframe, metadata, t_start, t_stop, samp
             # create a Neo SpikeTrain for each cluster label
             st = neo.SpikeTrain(
                 name = str(spike_label),
-                file_origin = abs_path(metadata, 'tridesclous_file'),
+                file_origin = _abs_path(metadata, 'tridesclous_file'),
                 channels = channels, # custom annotation
                 amplitude = None,    # custom annotation
                 times = t_start + sampling_period * df['index'].values,
@@ -335,9 +405,10 @@ def CreateNeoSpikeTrainsFromDataframe(dataframe, metadata, t_start, t_stop, samp
 
     return spiketrain_list
 
-def ApplyFilters(metadata, blk):
+def _apply_filters(metadata, blk):
     """
-
+    Apply filters specified in ``metadata`` the the signals in ``blk`` using
+    :func:`elephant.signal_processing.butter`.
     """
 
     if metadata['filters'] is not None:
@@ -367,9 +438,10 @@ def ApplyFilters(metadata, blk):
 
     return blk
 
-def RunAmplitudeDiscriminators(metadata, blk):
+def _run_amplitude_discriminators(metadata, blk):
     """
-
+    Run all amplitude discriminators for spike detection given in ``metadata``
+    on the signals in ``blk``.
     """
 
     spiketrain_list = []
@@ -390,15 +462,17 @@ def RunAmplitudeDiscriminators(metadata, blk):
             else:
 
                 sig = blk.segments[0].analogsignals[index]
-                st = DetectSpikes(sig, discriminator, epochs)
+                st = _detect_spikes(sig, discriminator, epochs)
                 spiketrain_list.append(st)
 
     return spiketrain_list
 
 
-def DetectSpikes(sig, discriminator, epochs):
+def _detect_spikes(sig, discriminator, epochs):
     """
-
+    Detect spikes in the amplitude window given by ``discriminator`` using
+    :func:`elephant.spike_train_generation.peak_detection`, and optionally
+    filter them by coincidence with epochs of a given name.
     """
 
     assert sig.name == discriminator['channel'], 'sig name "{}" does not match amplitude discriminator channel "{}"'.format(sig.name, discriminator['channel'])
@@ -412,8 +486,8 @@ def DetectSpikes(sig, discriminator, epochs):
     else:
         raise ValueError('amplitude discriminator must have two nonnegative thresholds or two nonpositive thresholds: {}'.format(discriminator))
 
-    spikes_crossing_min = elephant.spike_train_generation.peak_detection(sig, min_threshold*pq.uV, sign, 'raw')
-    spikes_crossing_max = elephant.spike_train_generation.peak_detection(sig, max_threshold*pq.uV, sign, 'raw')
+    spikes_crossing_min = elephant.spike_train_generation.peak_detection(sig, pq.Quantity(min_threshold, discriminator['units']), sign, 'raw')
+    spikes_crossing_max = elephant.spike_train_generation.peak_detection(sig, pq.Quantity(max_threshold, discriminator['units']), sign, 'raw')
     if sign == 'above':
         spikes_between_min_and_max = np.setdiff1d(spikes_crossing_min, spikes_crossing_max)
     elif sign == 'below':
