@@ -1,129 +1,53 @@
 # -*- coding: utf-8 -*-
 """
+The :mod:`neurotic.gui.config` module implements a class for configuring and
+launching ephyviewer for a loaded dataset.
 
+.. autoclass:: EphyviewerConfigurator
+   :members:
 """
 
-import os
-from collections import OrderedDict
-from functools import wraps
+import re
 
 import numpy as np
+import pandas as pd
 import quantities as pq
 import ephyviewer
-import ipywidgets
 
-from ..datasets import abs_path
-from ..utils import EstimateVideoJumpTimes, NeoEpochToDataFrame
-from ..gui.epochencoder import MyWritableEpochSource
+from ..datasets.metadata import _abs_path
+from ..gui.epochencoder import NeuroticWritableEpochSource
 
 pq.mN = pq.UnitQuantity('millinewton', pq.N/1e3, symbol = 'mN');  # define millinewton
 
-def _fix_FrameGrabber_set_file(set_file_func):
+
+class EphyviewerConfigurator():
     """
-    For some video files, ephyviewer sets the incorrect start_time for its
-    FrameGrabber when the set_file method is called. This monkey patch works
-    around the problem by resetting start_time to 0 after set_file is called.
+    A class for launching ephyviewer for a dataset with configurable viewers.
+
+    At initialization, invalid viewers are automatically disabled (e.g., the
+    video viewer is disabled if ``video_file`` is not given in ``metadata``).
+    Viewers can be hidden or shown before launch using the built-in methods.
+    Valid viewer names are:
+
+        * ``traces``
+        * ``traces_rauc``
+        * ``freqs``
+        * ``spike_trains``
+        * ``epochs``
+        * ``epoch_encoder``
+        * ``video``
+        * ``event_list``
+        * ``data_frame``
+
+    :meth:`launch_ephyviewer` is provided for starting a new Qt app and
+    launching the ephyviewer main window all at once.
+    :meth:`create_ephyviewer_window` generates just the ephyviewer window
+    and should be used if there is already a Qt app running.
     """
-    @wraps(set_file_func)
-    def wrapper(*args, **kwargs):
-        result = set_file_func(*args, **kwargs)
-        frame_grabber = args[0]
-        frame_grabber.start_time = 0
-        return result
-    return wrapper
-ephyviewer.FrameGrabber.set_file = _fix_FrameGrabber_set_file(ephyviewer.FrameGrabber.set_file)
-
-def defaultKeepSignal(sig):
-    """
-
-    """
-    return (not sig.name.startswith('Analog Input #')) and (sig.name != 'Clock')
-
-def defaultUnits(sig):
-    """
-
-    """
-
-    mapping = {
-        'V': 'uV', # convert voltages to microvolts
-        'N': 'mN', # convert forces to millinewtons
-    }
-    mapping = {pq.Quantity(1, k).dimensionality.simplified: v for k, v in mapping.items()}
-    return mapping.get(sig.units.dimensionality.simplified, sig.units)
-
-def defaultYLim(sig):
-    """
-
-    """
-
-    mapping = {
-        'V': [-120, 120], # plot range for voltages
-        'N': [ -10, 300], # plot range for forces
-    }
-    mapping = {pq.Quantity(1, k).dimensionality.simplified: v for k, v in mapping.items()}
-    return mapping.get(sig.units.dimensionality.simplified, [-1, 1])
-
-def setDefaultsForPlots(metadata, blk):
-    """
-
-    """
-
-    sigs = blk.segments[0].analogsignals
-    signalNameToIndex = {sig.name:i for i, sig in enumerate(sigs)}
-
-    if metadata['plots'] is None:
-        metadata['plots'] = [{'channel': sig.name} for sig in sigs if defaultKeepSignal(sig)]
-
-    plots = []
-    for plot in metadata['plots']:
-        index = signalNameToIndex.get(plot['channel'], None)
-        if index is None:
-            print('Warning: removing plot with channel name "{}" because channel was not found in blk!'.format(plot['channel']))
-        else:
-            plot['index'] = index
-            plot.setdefault('units',  defaultUnits(sigs[index]))
-            plot.setdefault('ylim',   defaultYLim(sigs[index]))
-            plot.setdefault('ylabel', sigs[index].name)
-            plots.append(plot)
-    metadata['plots'] = plots
-
-    return metadata['plots']
-
-class EphyviewerConfigurator(ipywidgets.HBox):
-    """
-
-    """
-
-    _toggle_button_defaults = OrderedDict([
-        ('traces',        {'value': True,  'icon': 'line-chart',   'description': 'Traces'}),
-        ('traces_rauc',   {'value': False, 'icon': 'area-chart',   'description': 'RAUC'}),
-        # ('freqs',         {'value': False, 'icon': 'wifi',         'description': 'Frequencies'}),
-        ('spike_trains',  {'value': True,  'icon': 'barcode',      'description': 'Spike Trains'}),
-        ('epochs',        {'value': True,  'icon': 'align-left',   'description': 'Read-Only Epochs'}),
-        ('epoch_encoder', {'value': True,  'icon': 'align-left',   'description': 'Epoch Encoder'}),
-        ('video',         {'value': True,  'icon': 'youtube-play', 'description': 'Video'}),
-        ('event_list',    {'value': True,  'icon': 'list',         'description': 'Events'}),
-        ('data_frame',    {'value': False, 'icon': 'table',        'description': 'Annotation Table'}),
-    ])
-
-    themes = {}
-    themes['original'] = None # special keyword to use ephyviewer's defaults
-    themes['light'] = {
-        'cmap': 'Dark2', # dark traces
-        'background_color': '#F0F0F0', # light gray
-        'vline_color': '#000000AA', # transparent black
-        'label_fill_color': '#DDDDDDDD', # transparent light gray
-    }
-    themes['dark'] = {
-        'cmap': 'Accent', # light traces
-        'background_color': 'k', # black
-        'vline_color': '#FFFFFFAA', # transparent white
-        'label_fill_color': '#222222DD', # transparent dark gray
-    }
 
     def __init__(self, metadata, blk, rauc_sigs = None, lazy = False):
         """
-
+        Initialize a new EphyviewerConfigurator.
         """
 
         self.metadata = metadata
@@ -131,50 +55,61 @@ class EphyviewerConfigurator(ipywidgets.HBox):
         self.rauc_sigs = rauc_sigs
         self.lazy = lazy
 
-        # initialize the box
-        ipywidgets.HBox.__init__(self)
+        self.viewer_settings = {
+            'traces':        {'show': True, 'disabled': False, 'reason': ''},
+            'traces_rauc':   {'show': False, 'disabled': False, 'reason': ''},
+            'freqs':         {'show': False, 'disabled': True, 'reason': 'Experimental and computationally expensive'},
+            'spike_trains':  {'show': True, 'disabled': False, 'reason': ''},
+            'epochs':        {'show': True, 'disabled': False, 'reason': ''},
+            'epoch_encoder': {'show': True, 'disabled': False, 'reason': ''},
+            'video':         {'show': True, 'disabled': False, 'reason': ''},
+            'event_list':    {'show': True, 'disabled': False, 'reason': ''},
+            'data_frame':    {'show': False, 'disabled': False, 'reason': ''},
+        }
 
-        # create buttons for controlling which elements to show
-        self.controls = OrderedDict()
-        for name, kwargs in self._toggle_button_defaults.items():
-            self.controls[name] = ipywidgets.ToggleButton(**kwargs)
-        controls_vbox = ipywidgets.VBox(list(self.controls.values()))
+        self.themes = {}
+        self.themes['original'] = None # special keyword to use ephyviewer's defaults
+        self.themes['light'] = {
+            'cmap': 'Dark2', # dark traces
+            'background_color': '#F0F0F0', # light gray
+            'vline_color': '#000000AA', # transparent black
+            'label_fill_color': '#DDDDDDDD', # transparent light gray
+        }
+        self.themes['dark'] = {
+            'cmap': 'Accent', # light traces
+            'background_color': 'k', # black
+            'vline_color': '#FFFFFFAA', # transparent white
+            'label_fill_color': '#222222DD', # transparent dark gray
+        }
 
-        # permanently disable controls for which inputs are missing
+        # hide and disable viewers for which inputs are missing
         if not self.rauc_sigs:
-            self.controls['traces_rauc'].value = False
-            self.controls['traces_rauc'].disabled = True
-            self.controls['traces_rauc'].tooltip = 'Cannot enable because rauc_sigs is empty'
+            self.viewer_settings['traces_rauc']['show'] = False
+            self.viewer_settings['traces_rauc']['disabled'] = True
+            self.viewer_settings['traces_rauc']['reason'] = 'Cannot enable because rauc_sigs is empty'
         if not self.blk.segments[0].spiketrains:
-            self.controls['spike_trains'].value = False
-            self.controls['spike_trains'].disabled = True
-            self.controls['spike_trains'].tooltip = 'Cannot enable because there are no spike trains'
+            self.viewer_settings['spike_trains']['show'] = False
+            self.viewer_settings['spike_trains']['disabled'] = True
+            self.viewer_settings['spike_trains']['reason'] = 'Cannot enable because there are no spike trains'
         if not [ep for ep in self.blk.segments[0].epochs if ep.size > 0 and '(from epoch encoder file)' not in ep.labels]:
-            self.controls['epochs'].value = False
-            self.controls['epochs'].disabled = True
-            self.controls['epochs'].tooltip = 'Cannot enable because there are no read-only epochs'
-            self.controls['data_frame'].value = False
-            self.controls['data_frame'].disabled = True
-            self.controls['data_frame'].tooltip = 'Cannot enable because there are no read-only epochs'
+            self.viewer_settings['epochs']['show'] = False
+            self.viewer_settings['epochs']['disabled'] = True
+            self.viewer_settings['epochs']['reason'] = 'Cannot enable because there are no read-only epochs'
+            self.viewer_settings['data_frame']['show'] = False
+            self.viewer_settings['data_frame']['disabled'] = True
+            self.viewer_settings['data_frame']['reason'] = 'Cannot enable because there are no read-only epochs'
             if not [ev for ev in self.blk.segments[0].events if ev.size > 0]:
-                self.controls['event_list'].value = False
-                self.controls['event_list'].disabled = True
-                self.controls['event_list'].tooltip = 'Cannot enable because there are no read-only epochs or events'
+                self.viewer_settings['event_list']['show'] = False
+                self.viewer_settings['event_list']['disabled'] = True
+                self.viewer_settings['event_list']['reason'] = 'Cannot enable because there are no read-only epochs or events'
         if not self.metadata['epoch_encoder_file']:
-            self.controls['epoch_encoder'].value = False
-            self.controls['epoch_encoder'].disabled = True
-            self.controls['epoch_encoder'].tooltip = 'Cannot enable because epoch_encoder_file is not set'
+            self.viewer_settings['epoch_encoder']['show'] = False
+            self.viewer_settings['epoch_encoder']['disabled'] = True
+            self.viewer_settings['epoch_encoder']['reason'] = 'Cannot enable because epoch_encoder_file is not set'
         if not self.metadata['video_file']:
-            self.controls['video'].value = False
-            self.controls['video'].disabled = True
-            self.controls['video'].tooltip = 'Cannot enable because video_file is not set'
-
-        # create the launch button
-        self.launch_button = ipywidgets.Button(icon='rocket', description='Launch', layout=ipywidgets.Layout(height='auto'))
-        self.launch_button.on_click(self._on_launch_clicked)
-
-        # populate the box
-        self.children = [controls_vbox, self.launch_button]
+            self.viewer_settings['video']['show'] = False
+            self.viewer_settings['video']['disabled'] = True
+            self.viewer_settings['video']['reason'] = 'Cannot enable because video_file is not set'
 
         # warn about potential video sync problems
         if metadata['video_file'] is not None and metadata['video_offset'] is None:
@@ -182,7 +117,7 @@ class EphyviewerConfigurator(ipywidgets.HBox):
             print('data because video_offset is unspecified! Consider adding')
             print('it to your metadata.')
         if metadata['video_file'] is not None and metadata['video_jumps'] is None:
-            approx_video_jumps = EstimateVideoJumpTimes(blk)
+            approx_video_jumps = _estimate_video_jump_times(blk)
             if approx_video_jumps:
                 print('WARNING: It seems that AxoGraph was paused at least once')
                 print('during data acquisition, but video_jumps is unspecified.')
@@ -199,49 +134,75 @@ class EphyviewerConfigurator(ipywidgets.HBox):
 
     def is_enabled(self, name):
         """
-
+        Return whether the viewer ``name`` is enabled.
         """
-        if name in self.controls:
-            return self.controls[name].value
+        if name in self.viewer_settings:
+            return not self.viewer_settings[name]['disabled']
         else:
             return False
 
     def enable(self, name):
         """
-
+        Enable the viewer ``name``.
         """
-        self.controls[name].value = True
+        if name in self.viewer_settings:
+            self.viewer_settings[name]['disabled'] = False
 
     def disable(self, name):
         """
-
+        Disable the viewer ``name``.
         """
-        self.controls[name].value = False
+        if name in self.viewer_settings:
+            self.viewer_settings[name]['disabled'] = True
 
-    def enable_all(self):
+    def is_shown(self, name):
         """
+        Return whether the viewer ``name`` is shown.
+        """
+        if name in self.viewer_settings:
+            return self.viewer_settings[name]['show']
+        else:
+            return False
 
+    def show(self, name):
         """
-        for name in self.controls:
-            if not self.controls[name].disabled:
-                self.enable(name)
+        Show the viewer ``name``.
+        """
+        if name in self.viewer_settings:
+            if not self.viewer_settings[name]['disabled']:
+                self.viewer_settings[name]['show'] = True
+            else:
+                print(self.viewer_settings[name]['reason'])
+        else:
+            print(f'"{name}" is not a viewer in viewer_settings')
 
-    def disable_all(self):
+    def hide(self, name):
         """
+        Hide the viewer ``name``.
+        """
+        if name in self.viewer_settings:
+            self.viewer_settings[name]['show'] = False
+        else:
+            print(f'"{name}" is not a viewer in viewer_settings')
 
+    def show_all(self):
         """
-        for name in self.controls:
-            self.disable(name)
+        Show all viewers.
+        """
+        for name in self.viewer_settings:
+            if not self.viewer_settings[name]['disabled']:
+                self.show(name)
 
-    def _on_launch_clicked(self, button):
+    def hide_all(self):
         """
-
+        Hide all viewers.
         """
-        self.launch_ephyviewer()
+        for name in self.viewer_settings:
+            self.hide(name)
 
     def launch_ephyviewer(self, theme='light', support_increased_line_width=False):
         """
-
+        Start a Qt app and create an ephyviewer window.
         """
 
         app = ephyviewer.mkQApp()
@@ -251,7 +212,7 @@ class EphyviewerConfigurator(ipywidgets.HBox):
 
     def create_ephyviewer_window(self, theme='light', support_increased_line_width=False):
         """
-
+        Load data into each ephyviewer viewer and return the main window.
         """
 
         ########################################################################
@@ -281,12 +242,15 @@ class EphyviewerConfigurator(ipywidgets.HBox):
             play_interval = 0.1, # refresh period in seconds
         )
         win.setWindowTitle(self.metadata['key'])
-        win.setWindowIcon(ephyviewer.QT.QIcon(':/soundwave.png'))
+        win.setWindowIcon(ephyviewer.QT.QIcon(':/neurotic-logo-150.png'))
+
+        # delete on close so that memory and file resources are released
+        win.setAttribute(ephyviewer.QT.WA_DeleteOnClose, True)
 
         ########################################################################
         # PREPARE TRACE PARAMETERS
 
-        setDefaultsForPlots(self.metadata, self.blk)
+        _set_defaults_for_plots(self.metadata, self.blk)
 
         plotNameToIndex = {p['channel']:i for i, p in enumerate(self.metadata['plots'])}
 
@@ -313,12 +277,14 @@ class EphyviewerConfigurator(ipywidgets.HBox):
         ########################################################################
         # TRACES WITH SCATTER PLOTS
 
-        if self.is_enabled('traces'):
+        if self.is_shown('traces'):
 
             if self.lazy:
                 import neo
-                neorawioclass = neo.rawio.get_rawio_class(abs_path(self.metadata, 'data_file'))
-                neorawio = neorawioclass(abs_path(self.metadata, 'data_file'))
+                neorawioclass = neo.rawio.get_rawio_class(_abs_path(self.metadata, 'data_file'))
+                if neorawioclass is None:
+                    raise ValueError('This file type cannot be read with fast loading (lazy=True): {}'.format(_abs_path(self.metadata, 'data_file')))
+                neorawio = neorawioclass(_abs_path(self.metadata, 'data_file'))
                 neorawio.parse_header()
 
                 # Intan-specific tricks
@@ -399,7 +365,7 @@ class EphyviewerConfigurator(ipywidgets.HBox):
         ########################################################################
         # TRACES OF RAUC
 
-        if self.is_enabled('traces_rauc') and self.rauc_sigs is not None:
+        if self.is_shown('traces_rauc') and self.rauc_sigs is not None:
 
             sig_rauc_source = ephyviewer.InMemoryAnalogSignalSource(
                 signals = np.concatenate([self.rauc_sigs[p['index']].as_array() for p in self.metadata['plots']], axis = 1),
@@ -442,7 +408,7 @@ class EphyviewerConfigurator(ipywidgets.HBox):
         ########################################################################
         # FREQUENCY (EXPERIMENTAL AND COMPUTATIONALLY EXPENSIVE!)
 
-        if self.is_enabled('freqs'):
+        if self.is_shown('freqs'):
 
             freq_view = ephyviewer.TimeFreqViewer(source = trace_view.source, name = 'timefreqs')
 
@@ -474,7 +440,7 @@ class EphyviewerConfigurator(ipywidgets.HBox):
         ########################################################################
         # SPIKE TRAINS
 
-        if self.is_enabled('spike_trains') and sources['spike'][0].nb_channel > 0:
+        if self.is_shown('spike_trains') and sources['spike'][0].nb_channel > 0:
 
             spike_train_view = ephyviewer.SpikeTrainViewer(source = sources['spike'][0], name = 'spiketrains')
             win.add_view(spike_train_view)
@@ -490,7 +456,7 @@ class EphyviewerConfigurator(ipywidgets.HBox):
         ########################################################################
         # EPOCHS
 
-        if self.is_enabled('epochs') and sources['epoch'][0].nb_channel > 0:
+        if self.is_shown('epochs') and sources['epoch'][0].nb_channel > 0:
 
             epoch_view = ephyviewer.EpochViewer(source = sources['epoch'][0], name = 'epochs')
             win.add_view(epoch_view)
@@ -506,11 +472,21 @@ class EphyviewerConfigurator(ipywidgets.HBox):
         ########################################################################
         # EPOCH ENCODER
 
-        if self.is_enabled('epoch_encoder') and self.metadata['epoch_encoder_file'] is not None:
+        if self.is_shown('epoch_encoder') and self.metadata['epoch_encoder_file'] is not None:
 
-            writable_epoch_source = MyWritableEpochSource(
-                filename = abs_path(self.metadata, 'epoch_encoder_file'),
-                possible_labels = self.metadata['epoch_encoder_possible_labels'],
+            possible_labels = self.metadata['epoch_encoder_possible_labels']
+
+            # append labels found in the epoch encoder file but not in the
+            # epoch_encoder_possible_labels list, preserving the original
+            # ordering of epoch_encoder_possible_labels
+            labels_from_file = [ep.name for ep in seg.epochs if len(ep.times) > 0 and '(from epoch encoder file)' in ep.labels]
+            for label in labels_from_file:
+                if label not in possible_labels:
+                    possible_labels.append(label)
+
+            writable_epoch_source = NeuroticWritableEpochSource(
+                filename = _abs_path(self.metadata, 'epoch_encoder_file'),
+                possible_labels = possible_labels,
             )
 
             epoch_encoder = ephyviewer.EpochEncoder(source = writable_epoch_source, name = 'epoch encoder')
@@ -527,14 +503,19 @@ class EphyviewerConfigurator(ipywidgets.HBox):
         ########################################################################
         # VIDEO
 
-        if self.is_enabled('video') and self.metadata['video_file'] is not None:
+        if self.is_shown('video') and self.metadata['video_file'] is not None:
 
-            video_source = ephyviewer.MultiVideoFileSource(video_filenames = [abs_path(self.metadata, 'video_file')])
+            video_source = ephyviewer.MultiVideoFileSource(video_filenames = [_abs_path(self.metadata, 'video_file')])
+
+            # some video files are loaded with an incorrect start time, so
+            # reset video start to zero
+            video_source.t_stops[0] -= video_source.t_starts[0]
+            video_source.t_starts[0] = 0
+
+            # apply the video_offset
             if self.metadata['video_offset'] is not None:
                 video_source.t_starts[0] += self.metadata['video_offset']
                 video_source.t_stops[0]  += self.metadata['video_offset']
-                video_source._t_start = max(min(video_source.t_starts), 0)
-                video_source._t_stop  = max(video_source.t_stops)
 
             # correct for videos that report frame rates that are too fast or
             # too slow compared to the clock on the data acquisition system
@@ -558,8 +539,10 @@ class EphyviewerConfigurator(ipywidgets.HBox):
                 video_source.video_times = [video_times]
                 video_source.t_starts[0] = min(video_times)
                 video_source.t_stops[0]  = max(video_times)
-                video_source._t_start = max(min(video_source.t_starts), 0)
-                video_source._t_stop  = max(video_source.t_stops)
+
+            # update the source-level times from the modified file-level times
+            video_source._t_start = max(min(video_source.t_starts), 0)
+            video_source._t_stop  = max(video_source.t_stops)
 
             video_view = ephyviewer.VideoViewer(source = video_source, name = 'video')
             if theme != 'original':
@@ -569,7 +552,7 @@ class EphyviewerConfigurator(ipywidgets.HBox):
         ########################################################################
         # EVENTS
 
-        if self.is_enabled('event_list') and sources['event'][0].nb_channel > 0:
+        if self.is_shown('event_list') and sources['event'][0].nb_channel > 0:
 
             event_list = ephyviewer.EventList(source = sources['event'][0], name = 'events')
             if 'video' in win.viewers:
@@ -580,8 +563,8 @@ class EphyviewerConfigurator(ipywidgets.HBox):
         ########################################################################
         # DATAFRAME
 
-        annotations_dataframe = NeoEpochToDataFrame(seg.epochs, exclude_epoch_encoder_epochs=True)
-        if self.is_enabled('data_frame') and len(annotations_dataframe) > 0:
+        annotations_dataframe = _neo_epoch_to_dataframe(seg.epochs, exclude_epoch_encoder_epochs=True)
+        if self.is_shown('data_frame') and len(annotations_dataframe) > 0:
 
             data_frame_view = ephyviewer.DataFrameView(source = annotations_dataframe, name = 'table')
             if 'events' in win.viewers:
@@ -603,3 +586,126 @@ class EphyviewerConfigurator(ipywidgets.HBox):
         win.set_xsize(self.metadata['t_width']) # seconds
 
         return win
+
+def _set_defaults_for_plots(metadata, blk):
+    """
+    Set defaults for plot channels, units, ylim, and ylabel if these
+    parameters are missing from ``metadata``.
+    """
+
+    sigs = blk.segments[0].analogsignals
+    signalNameToIndex = {sig.name:i for i, sig in enumerate(sigs)}
+
+    if metadata['plots'] is None:
+        metadata['plots'] = [{'channel': sig.name} for sig in sigs if _default_keep_signal(sig)]
+
+    plots = []
+    for plot in metadata['plots']:
+        index = signalNameToIndex.get(plot['channel'], None)
+        if index is None:
+            print('Warning: removing plot with channel name "{}" because channel was not found in blk!'.format(plot['channel']))
+        else:
+            plot['index'] = index
+            plot.setdefault('units',  _default_units(sigs[index]))
+            plot.setdefault('ylim',   _default_ylim(sigs[index]))
+            plot.setdefault('ylabel', sigs[index].name)
+            plots.append(plot)
+    metadata['plots'] = plots
+
+    return metadata['plots']
+
+def _default_keep_signal(sig):
+    """
+    If ``plots`` is not specified in ``metadata``, this function determines
+    which channels are plotted by default.
+    """
+    return (not sig.name.startswith('Analog Input #')) and (sig.name != 'Clock')
+
+def _default_units(sig):
+    """
+    If ``plots`` is missing ``units`` in ``metadata``, this function determines
+    default units.
+    """
+
+    mapping = {
+        'V': 'uV', # convert voltages to microvolts
+        'N': 'mN', # convert forces to millinewtons
+    }
+    mapping = {pq.Quantity(1, k).dimensionality.simplified: v for k, v in mapping.items()}
+    return mapping.get(sig.units.dimensionality.simplified, sig.units)
+
+def _default_ylim(sig):
+    """
+    If ``plots`` is missing ``ylim`` in ``metadata``, this function determines
+    default plot ranges.
+    """
+
+    mapping = {
+        'V': [-120, 120], # plot range for voltages
+        'N': [ -10, 300], # plot range for forces
+    }
+    mapping = {pq.Quantity(1, k).dimensionality.simplified: v for k, v in mapping.items()}
+    return mapping.get(sig.units.dimensionality.simplified, [-1, 1])
+
+def _neo_epoch_to_dataframe(neo_epochs, exclude_epoch_encoder_epochs=False):
+    """
+    Convert a list of Neo Epochs into a dataframe.
+    """
+
+    dtypes = {
+        'Start (s)':    float,
+        'End (s)':      float,
+        'Duration (s)': float,
+        'Type':         str,
+        'Label':        str,
+    }
+    columns = list(dtypes.keys())
+    df = pd.DataFrame(columns=columns)
+    for ep in neo_epochs:
+        if not exclude_epoch_encoder_epochs or '(from epoch encoder file)' not in ep.labels:
+            data = np.array([ep.times, ep.times+ep.durations, ep.durations, [ep.name]*len(ep), ep.labels]).T
+            df = df.append(pd.DataFrame(data, columns=columns), ignore_index=True)
+    return df.astype(dtype=dtypes).sort_values(['Start (s)', 'End (s)', 'Type', 'Label']).reset_index(drop=True)
+
+def _estimate_video_jump_times(blk):
+    """
+    Estimate how much time to skip in video playback if AxoGraph was temporarily
+    paused during data acquisition while the video continued to record. Returns
+    a list of ordered pairs suitable for the video_jumps metadata parameter. The
+    returned stop times are exact, but pause durations have only whole-second
+    precision and should be manually refined by inspecting the video before
+    using.
+    """
+
+    if 'notes' not in blk.annotations:
+        return None
+    else:
+
+        # obtain approximate start and stop times according to AxoGraph notes
+        note_start_times = np.array([0], dtype=np.int)
+        note_stop_times = np.array([], dtype=np.int)
+        for note_line in blk.annotations['notes'].split('\n'):
+            m = re.match('\d\d\d: Start at (\d*) s', note_line)
+            if m:
+                note_start_times = np.append(note_start_times, int(m.group(1)))
+            m = re.match('\d\d\d: Stop at (\d*) s', note_line)
+            if m:
+                note_stop_times = np.append(note_stop_times, int(m.group(1)))
+
+        # calculate approximate pause durations
+        pause_durations = note_start_times[1:]-note_stop_times[:-1]
+
+        # obtain exact stop times (AxoGraph time, not video time)
+        event_stop_times = np.array([], dtype=np.float)
+        ev = next((ev for ev in blk.segments[0].events if ev.name == 'AxoGraph Tags'), None)
+        if ev is not None:
+            for time, label in zip(ev.times, ev.labels):
+                if label == 'Stop':
+                    event_stop_times = np.append(event_stop_times, time.magnitude)
+
+        # pair stop times with pause durations
+        video_jumps = []
+        for t, dur in zip(event_stop_times[:-1], pause_durations):
+            video_jumps.append([t, dur])
+
+        return video_jumps
